//这段代码是用于控制一个名为MyCobot的机械臂的基本操作，特别是关于其夹爪（gripper）的开合角度和速度。
//代码中使用了MyCobotBasic这个类，这个类很可能是一个封装了与MyCobot机械臂交互功能的库

#include <MyCobotBasic.h>   // 引入MyCobotBasic库，这个库包含了控制MyCobot机械臂所需的所有基本函数  
  
MyCobotBasic myCobot;  // 创建一个MyCobotBasic类的实例，用于后续控制机械臂  
  
// setup()函数在Arduino板子上电或重置后自动执行一次  
void setup()  
{  
    // 调用myCobot的setup()函数，这个函数通常用于初始化与机械臂的连接或配置  
    myCobot.setup(); // This api is required  
    delay(100); // 等待100毫秒，确保初始化完成  
    // 打开机械臂的电源  
    myCobot.powerOn(); // robot power on  
    delay(100); // 等待100毫秒，确保电源稳定  
}  
  
// loop()函数在setup()函数执行完后不断循环执行  
void loop()  
{  
    // 将夹爪设置到80度的位置，速度为50（这里的速度单位可能是百分比或其他，具体看文档）  
    myCobot.setGripperValue(80, 50); // let the gripper to 80°, speed 50  
    delay(500); // 等待500毫秒，让夹爪到达指定位置  
      
    // 将夹爪从80度调整到20度，速度仍为50  
    myCobot.setGripperValue(20, 50); // let the gripper to 20°, speed 50  
    delay(500); // 等待500毫秒  
      
    // 打开夹爪，速度设置为30  
    // 注意：这里的setGripperState函数可能是通过不同的参数来控制夹爪的开合状态，而不是直接设置角度  
    // 假设0表示打开，1表示关闭  
    myCobot.setGripperState(0, 30); // open the gripper, speed 30  
    delay(600); // 等待600毫秒，确保夹爪完全打开  
      
    // 关闭夹爪，速度同样为30  
    myCobot.setGripperState(1, 30); // close the gripper, speed 30  
    delay(600); // 等待600毫秒，确保夹爪完全关闭  
      
    // 这个LOOP循环会持续不断地执行上述步骤  
}
