#include <MyCobotBasic.h>  // 引入MyCobotBasic库
#include <vector>  // 引入标准库vector

#define SET_BASIC_OUT 0xa0  // 定义常量用于设置输出
#define GET_BASIC_IN 0xa1  // 定义常量用于获取输入

byte HEADER = 0xfe;  // 定义消息头
byte FOOTER = 0xfa;  // 定义消息尾

using namespace std;
vector<int> client_data;  // 存储从客户端接收的数据
vector<int> temp_data;  // 临时存储数据
vector<int> atom_data;  // 存储从Atom接收的数据
int count = -1, data = -1;  // 计数器和数据变量
int t_begin;  // 记录时间的变量

void setup() {
    Serial.begin(115200);  // 初始化串口通信，波特率为115200
    /* 如果你使用其他板子，请检查官方文档，确认串口是否支持 Serial1。
     * 例如，M5 的串口是 Serial1（所以将下面的 Serial1 替换为 Serial1）。*/
    Serial1.begin(1000000);  // 初始化Serial1，波特率为1000000
    delay(100);

    MyCobotBasic myCobot;  // 创建MyCobotBasic对象
    delay(500);

    myCobot.powerOn();  // 给MyCobot上电
    delay(500);

    pinMode(15, OUTPUT);  // 设置引脚15为输出
    pinMode(5, OUTPUT);  // 设置引脚5为输出
    pinMode(2, OUTPUT);  // 设置引脚2为输出
}

void loop() {
    ClientDataToAtom();  // 处理从客户端到Atom的数据
    AtomDataToClient();  // 处理从Atom到客户端的数据
    OtherMsgToClient();  // 处理其他消息并发送给客户端
    ClearData();  // 清除数据
}

/*
 * 函数: 检查串口WiFi或蓝牙返回消息的头部是否是 fe fe
 */
bool checkHeader(vector<int> &v_data) {
    if (v_data[0] == HEADER && v_data[1] == HEADER)
        return true;
    return false;
}

bool HandleStickyPackets(vector<int> &v_data) {
    // 获取协议的长度以确定终止符 fa
    vector<int>::iterator it_vdata = v_data.begin() + 2 + v_data[2];
    int len = v_data.size();
    // 首先判断长度是否小于5，指令长度至少为5--> 处理数据
    if (len >= 5 && len <= (v_data[2] + 3)) {
        if (*it_vdata == FOOTER) {
            v_data.erase(it_vdata + 1, v_data.end());
            return true;
        }
    } 
    return false;
}

/*
 * 函数: 处理一些Atom没有处理的客户端消息
 */
bool HandleOtherMsg(vector<int> &v_data) {
    bool flag = false;
    switch (v_data[2]) {
        case 4:
            switch (v_data[3]) {
                case SET_BASIC_OUT: {
                    //0xfe 0xfe 0x04 0xa0 pin_no pin_data 0xfa
                    byte pin_no = v_data[4];
                    pinMode(pin_no, OUTPUT);
                    delay(5);
                    bool pin_data = v_data[5];
                    digitalWrite(pin_no, pin_data);
                }
                break;
            }
            break;
        case 3:
            switch (v_data[3]) {
                case GET_BASIC_IN: {
                    byte pin_no = v_data[4];
                    pinMode(pin_no, INPUT);
                    delay(5);
                    bool pin_state = digitalRead(pin_no);
                    delay(5);
                    v_data[2] = 0x04;
                    v_data.insert(v_data.end() - 1, pin_state);
                    flag = true;
                }
                break;
            }
            break;
    }
    return flag;
}

/*
 * 函数: 将Atom返回的消息发送给用户
 */
void SendDataToUser(vector<int> &v_data) {
    if (checkHeader(v_data) && HandleStickyPackets(v_data)) {
        if (HandleOtherMsg(v_data)) {
            for (int i = 0; i < v_data.size(); i++) {
                Serial.write(v_data[i]);
            }
        }
        v_data.clear();
    }
    return;
}

void ClientDataToAtom() {
    if (Serial.available() > 0) {
        count = 1; 
        while (Serial.available() > 0) {
            data = Serial.read();
            Serial1.write(data);
            client_data.push_back(data);
        }
        t_begin = millis();
    }
}

void AtomDataToClient() {
    int temp;
    bool flag = false;
    if (Serial1.available() > 0) {
        count = 0;
        while (Serial1.available() > 0) {
            data = Serial1.read();
            if (data == HEADER && (temp = Serial1.read()) == HEADER) {
                Serial.write(temp);
                flag = true;
            }
            if (flag) {
                Serial.write(data);
            }
        }
    }
}

void OtherMsgToClient() {
    if (count == 1 && ((millis() - t_begin) > 10)) {
        count = 0;
        for (int i = 0; i < client_data.size(); i++) {
            temp_data.push_back(client_data[i]);
        }
        SendDataToUser(temp_data); 
    }
}

void ClearData() {
    if (count == 0){
        client_data.clear();
    }
}
