//这段代码是用于控制一个名为MyCobot的机械臂的，特别是关于其关节角度和末端执行器（如抓手或工具）在空间中的位置控制。
//代码中使用了MyCobotBasic这个类，该类封装了与MyCobot机械臂交互的功能。
//此外，还定义了一个Coords结构体（或类，具体取决于MyCobotBasic库的实现），用于存储机械臂末端执行器在空间中的坐标。


#include <MyCobotBasic.h>  // 引入MyCobotBasic库，该库包含了控制MyCobot机械臂所需的所有基本函数  
 
MyCobotBasic myCobot;   // 创建MyCobotBasic类的实例，用于后续控制机械臂  
  
// 定义一个Coords类型的变量，用于存储机械臂末端执行器在空间中的坐标  
// 假设Coords是一个结构体或类，具有六个成员（可能是x, y, z, rx, ry, rz或类似的表示）  
// 这里初始化为特定的坐标值  
Coords coords = {194.700, -67.400, 131.300, -177.260, -3.760, -110.880};  
  
// setup()函数在Arduino板子上电或重置后自动执行一次  
void setup()  
{  
    // 调用myCobot的setup()函数，这个函数通常用于初始化与机械臂的连接或配置  
    myCobot.setup(); // This api is required  
    delay(100); // 等待100毫秒，确保初始化完成  
      
    // 打开机械臂的电源  
    myCobot.powerOn(); // robot power on  
    delay(100); // 等待100毫秒，确保电源稳定  
      
    // 将机械臂的关节角度设置为初始姿态，速度为50（速度单位可能是百分比或其他）  
    // 注意：这里假设{0, -10, -123, 45, 0, 20}是六个关节的初始角度值  
    myCobot.writeAngles({0, -10, -123, 45, 0, 20}, 50); // Initial attitude of coordinate control  
    delay(6000); // 等待6000毫秒（6秒），可能是为了观察机械臂移动到初始姿态或进行调试  
}  
  
// loop()函数在setup()函数执行完后不断循环执行  
void loop()  
{  
    // 控制机械臂在某一轴（这里假设是Axis枚举中的第一个轴，可能是X轴）上移动30mm，速度为30  
    // 注意：(Axis)1可能是一个类型转换，但通常轴的控制更直接地通过x, y, z等参数进行，除非(Axis)是特定于库的  
    // 这里我们假设它有效，但具体取决于MyCobotBasic库的实现  
    myCobot.writeCoord((Axis)1, 30, 30); // let x to 30mm  
    delay(300); // 等待300毫秒，让机械臂移动到指定位置  
      
    // 将机械臂的末端执行器移动到coords变量指定的坐标位置，速度为30  
    // 这将覆盖之前对单轴的单独设置  
    myCobot.writeCoords(coords, 30); // Multi-coordinate control  
    delay(5000); // 等待5000毫秒（5秒），可能是为了观察机械臂的移动或进行调试  
      
    // 注意：由于loop函数是不断循环的，所以上述动作会不断重复执行  
    // 如果想要执行一次后停止，可以在执行完毕后添加某种形式的退出条件（例如，使用一个标志变量）  
}
