//这段代码是用于控制一个名为MyCobot的机械臂的，特别是关于其抓手（gripper）的打开和关闭操作，以及抓手的角度调整。
//代码中使用了MyCobotBasic这个类，该类提供了与MyCobot机械臂交互的接口。
#include <MyCobotBasic.h> // 引入MyCobotBasic库，以便使用该类提供的函数  
  
MyCobotBasic myCobot; // 创建MyCobotBasic类的实例，用于后续控制机械臂  
  
// setup()函数在Arduino板子上电或重置后自动执行一次  
void setup()  
{  
    myCobot.setup(); // 调用myCobot的setup()函数，这个函数通常用于初始化与机械臂的连接或配置  
    delay(100); // 等待100毫秒，确保初始化完成  
    myCobot.powerOn(); // 打开机械臂的电源  
    delay(100); // 等待100毫秒，确保电源稳定  
}  
  
// loop()函数在setup()函数执行完后不断循环执行  
void loop()  
{  
    // 将抓手的角度设置为80°，速度为50（速度单位可能是百分比或其他，具体取决于库的实现）  
    myCobot.setGripperValue(80, 50); // let the gripper to 80°, speed 50  
    delay(500); // 等待500毫秒，让抓手移动到指定角度  
      
    // 将抓手的角度调整为20°，速度仍为50  
    myCobot.setGripperValue(20, 50); // let the gripper to 20°, speed 50  
    delay(500); // 再次等待500毫秒  
      
    // 打开抓手，速度为30  
    // 注意：这里的setGripperState函数可能不是直接设置角度，而是控制抓手的开/关状态  
    // 第一个参数0可能表示“打开”状态，第二个参数是速度  
    myCobot.setGripperState(0, 30); // open the gripper, speed 30  
    delay(600); // 等待600毫秒，确保抓手完全打开  
      
    // 关闭抓手，速度仍为30  
    // 第一个参数1可能表示“关闭”状态  
    myCobot.setGripperState(1, 30); // close the gripper, speed 30  
    delay(600); // 等待600毫秒，确保抓手完全关闭  
      
    // 由于loop函数是不断循环的，所以上述动作会不断重复执行  
    // 如果想要执行一次后停止，可以在执行完毕后添加某种形式的退出条件（例如，使用一个标志变量）  
}
