//这段代码是用于控制一个名为MyCobot的机械臂的，特别是关于其关节（joints）的角度控制和机械臂的电源管理。
//代码中使用了MyCobotBasic这个类，该类提供了与MyCobot机械臂交互的接口。
#include <MyCobotBasic.h> // 引入MyCobotBasic库，以便使用该类提供的函数  
  
MyCobotBasic myCobot; // 创建MyCobotBasic类的实例，用于后续控制机械臂  
Angles angles = {0, 0, 0, 0, 0, 0}; // 定义一个Angles类型的变量并初始化为0，用于存储关节角度，但这里未使用非零值  
  
// setup()函数在Arduino板子上电或重置后自动执行一次  
void setup()  
{  
    myCobot.setup(); // 调用myCobot的setup()函数，这个函数通常用于初始化与机械臂的连接或配置  
    delay(100); // 等待100毫秒，确保初始化完成  
    myCobot.powerOn(); // 打开机械臂的电源  
    delay(100); // 等待100毫秒，确保电源稳定  
}  
  
// loop()函数在setup()函数执行完后不断循环执行  
void loop()  
{  
    // 控制机械臂的第一个关节（J1）移动到100°，速度为30（速度单位可能是百分比或其他）  
    myCobot.writeAngle((Joint)1, 100, 30); // Single joint control，J1 to 100°, speed 30  
    // 注意：这里的(Joint)1是一个类型转换，假设Joint是一个枚举类型，用于指定关节编号  
    delay(200); // 等待200毫秒，让关节移动到指定位置  
      
    // 将所有关节的角度设置为angles变量中的值（在这个例子中，所有关节都被设置为0°），速度为50  
    // 尽管angles变量的值都是0，但这个调用展示了如何同时控制多个关节  
    myCobot.writeAngles(angles, 50); // All joints to zero, speed 50  
    delay(5000); // 等待5000毫秒（5秒），可能是为了观察机械臂的移动效果或进行调试  
      
    // 由于loop函数是不断循环的，所以上述动作会不断重复执行  
    // 如果想要执行一次后停止，可以在执行完毕后添加某种形式的退出条件（例如，使用一个标志变量）  
}
